#include <gmp.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

// Define these function's to make it more readable
#define mpz_rshift(x, y) mpz_tdiv_q_2exp(x, y, 1)
#define mpz_lshift(x, y) mpz_mul_2exp(x, y, 1)

#define mpz_dec(x) mpz_sub_ui(x, x, 1);
#define mpz_inc(x) mpz_add_ui(x, x, 1);

int main (int argc, char *argv[])
{
	gmp_randstate_t state;
	mpz_t one, size, n, e, p, q, pq, d, tap, m;
	
	// Initialze and set default values for 'e' and 'n'
	mpz_init_set_si(e, 65537);
	mpz_init_set_str(n, "aba7f0b8c7feb6e33438af5c25c67389eaf4d73f80cd0a37922493431cde03b34da448bdb05387cf7a8c59ee12d9613429a2b07ea385752f079892da1ae76c2b158f2d7169aae066432fe44f797df39dd6a0d7b2e2091281b30efac247c51576ebc93ec456de2e27d36b713844336b65af67ee58e6107a6a1deb954a91095295", 16);
	
	// Loop through arguments
	while(--argc)
	{
		argv++;
		
		// If the user has defined a value for 'e', use it rather than default
		if(strcmp(*argv, "-e"))
		{
			argc--;
			argv++;
			mpz_set_str(e, *argv, 16);
		}
		
		// If the user has requested help on the command, give it to them
		else if(strcmp(*argv, "-h") || strcmp(*argv, "--help"))
		{
			printf("Usage: client [-h] [-e e] [n]\n");
			return 0;
		}
		
		// If none of the above conditions are satisfied, assume the user entered a value for n
		else
		{
			argc--;
			argv++;
			mpz_set_str(n, *argv, 16);
		}
	}
	
	// Some error checking
	if(!mpz_cmp_si(e, -1) || !mpz_cmp_si(e, 0) || !mpz_cmp_si(n, -1) || !mpz_cmp_si(n, 0))
	{
		printf("::: ERROR: Invalid number\n");
		printf(":: Please enter a valid natural base 16 number without a trailing '0x'\n");
		return -1;
	}
	
	// Initialize all the other variables needed
	gmp_randinit_mt(state);
	mpz_init_set_si(size, 1300);
	mpz_init_set_si(one, 0);
	mpz_init(p);
	mpz_init(q);
	mpz_init(pq);
	mpz_init(d);
	mpz_init_set_str(tap, "0xa4800000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000", 16);
	mpz_init(m);
	
	// Set a random starting(not really sure of a value to put for the third argument)
	mpz_urandomm(p, state, size);
	
	// We don't need the random state anymore, so free the memory back up
	gmp_randclear(state);
	
	// Set 'm' for an initial test
	mpz_mod(m, n, p);
	
	while(mpz_cmp_si(m, 0))
	{
		// If 'p' is prime, right shift by one, otherwise do the same thing and then xor the number
		if(mpz_probab_prime_p(p, 5)) { mpz_rshift(p, p); }
		else
		{
			mpz_rshift(p, p);
			mpz_xor(p, p, tap);
		}
		gmp_printf(": p=%#Zx\r", p);
		mpz_mod(m, n, p);
	}
	
	// If we make it this far, 'p' and 'q' were found
	// So let's print the good news!
	printf("\n::: Found 'p' and 'q'!!!\n");
	printf(":: Calculating private key...\n");
	
	// Calculate the private key.
	mpz_tdiv_q(q, n, p);	// Divide 'n' by 'p' and store the result in 'q'. (The division truncates)
	mpz_sub_ui(pq, p, 1);	// Subtruct one from 'p' and store the result in 'pq'
	mpz_dec(q);				// Decrement 'q' by one
	mpz_mul(pq, pq, q);		// Get the result of '(q-1)(p-1)'
	mpz_inc(q);				// Increment 'q'
	mpz_invert(d, e, pq);	// Get the mod inverse of 'e' and 'pq' and store it in 'd'. (This get's the private key)
	
	// Finally, output all the values
	gmp_printf(": n = %#Zx\n", n);
	gmp_printf(": p = %#Zx\n", p);
	gmp_printf(": q = %#Zx\n", q);
	gmp_printf(": e = %#Zx\n", e);
	gmp_printf(": d = %#Zx\n", d);
	
	printf(":: Saving to file 'key.txt'...\n");
	
	FILE *fp;
	fp = fopen("key.txt", "w");
	
	// Check that we were able to open a file stream
	if(fp != NULL)
	{
		// If we were, start saving the data
		fprintf(fp, "n = ");
		mpz_out_str(fp, 16, n);
		
		fprintf(fp, "\np = ");
		mpz_out_str(fp, 16, p);
		
		fprintf(fp, "\nq = ");
		mpz_out_str(fp, 16, q);
		
		fprintf(fp, "\ne = ");
		mpz_out_str(fp, 16, e);
		
		fprintf(fp, "\nd = ");
		mpz_out_str(fp, 16, d);
		
		fclose(fp);
	}
	else
	{
		// Otherwise, alert the user that the program will not quit and
		// will stay open until they close it
		printf("::: ERROR: Unable to save to file!!!\n");
		printf(":: This program will stay open until explicitly closed.");
		
		// Yes, I know I could have done while(1); but I just felt like putting this in here
		__asm("loop:\nnop\njmp loop");
	}
	
	return 0;
}
